@@ -11,6 +11,18 @@
 // We want to keep cloudpinyin logic but don't call it.
 #include "../../modules/cloudpinyin/cloudpinyin_public.h"
 #include "config.h"
+#include "punctuation.h"
+#include <ctime>
+#include <fcitx-utils/keysymgen.h>
+#include <fcitx-utils/stringutils.h>
+#include <fcitx/candidatelist.h>
+#include <fcitx/event.h>
+#include <fcitx/userinterface.h>
+#include <memory>
+#include <string>
+#include <string_view>
+#include <unordered_map>
+#include <unordered_set>
 #ifdef FCITX_HAS_LUA
 #include "luaaddon_public.h"
 #endif
@@ -18,7 +30,6 @@
 #include "pinyinhelper_public.h"
 #include "punctuation_public.h"
 #include "spell_public.h"
-#include <boost/algorithm/string.hpp>
 #include <boost/iostreams/device/file_descriptor.hpp>
 #include <boost/iostreams/stream.hpp>
 #include <fcitx-config/iniparser.h>
@@ -35,6 +46,7 @@
 #include <fcitx/inputpanel.h>
 #include <fcitx/userinterfacemanager.h>
 #include <fcntl.h>
+#include <fmt/chrono.h>
 #include <fmt/format.h>
 #include <libime/core/historybigram.h>
 #include <libime/core/prediction.h>
@@ -54,14 +66,14 @@ FCITX_DEFINE_LOG_CATEGORY(pinyin, "pinyin");
 #define PINYIN_ERROR() FCITX_LOGC(pinyin, Error)
 
 bool consumePrefix(std::string_view &view, std::string_view prefix) {
-    if (boost::starts_with(view, prefix)) {
+    if (stringutils::startsWith(view, prefix)) {
         view.remove_prefix(prefix.size());
         return true;
     }
     return false;
 }
 
-enum class PinyinMode { Normal, StrokeFilter, ForgetCandidate };
+enum class PinyinMode { Normal, StrokeFilter, ForgetCandidate, Punctuation };
 
 class PinyinState : public InputContextProperty {
 public:
@@ -83,7 +95,7 @@ public:
 
     std::unique_ptr<EventSourceTime> cancelLastEvent_;
 
-    std::vector<std::string> predictWords_;
+    std::optional<std::vector<std::string>> predictWords_;
 
     int keyReleased_ = -1;
     int keyReleasedIndex_ = -2;
@@ -97,14 +109,17 @@ public:
     void select(InputContext *inputContext) const override {
         inputContext->commitString(word_);
         auto *state = inputContext->propertyFor(&engine_->factory());
-        state->predictWords_.push_back(word_);
+        if (!state->predictWords_) {
+            state->predictWords_.emplace();
+        }
+        auto &predictWords = *state->predictWords_;
+        predictWords.push_back(word_);
         // Max history size.
         constexpr size_t maxHistorySize = 5;
-        if (state->predictWords_.size() > maxHistorySize) {
-            state->predictWords_.erase(state->predictWords_.begin(),
-                                       state->predictWords_.begin() +
-                                           state->predictWords_.size() -
-                                           maxHistorySize);
+        if (predictWords.size() > maxHistorySize) {
+            predictWords.erase(predictWords.begin(), predictWords.begin() +
+                                                         predictWords.size() -
+                                                         maxHistorySize);
         }
         engine_->updatePredict(inputContext);
     }
@@ -113,11 +128,30 @@ public:
     std::string word_;
 };
 
-class StrokeCandidateWord : public CandidateWord {
+class PinyinAbstractExtraCandidateWordInterface {
+public:
+    explicit PinyinAbstractExtraCandidateWordInterface(CandidateWord &cand,
+                                                       int order)
+        : cand_(cand), order_(order) {}
+
+    virtual ~PinyinAbstractExtraCandidateWordInterface() = default;
+
+    int order() const { return order_; };
+    const CandidateWord &toCandidateWord() const { return cand_; }
+    CandidateWord &toCandidateWord() { return cand_; }
+
+private:
+    CandidateWord &cand_;
+    int order_;
+};
+
+class StrokeCandidateWord : public CandidateWord,
+                            public PinyinAbstractExtraCandidateWordInterface {
 public:
     StrokeCandidateWord(PinyinEngine *engine, std::string hz,
-                        const std::string &py)
-        : engine_(engine), hz_(std::move(hz)) {
+                        const std::string &py, int order)
+        : PinyinAbstractExtraCandidateWordInterface(*this, order),
+          engine_(engine), hz_(std::move(hz)) {
         if (py.empty()) {
             setText(Text(hz_));
         } else {
@@ -135,6 +169,52 @@ private:
     std::string hz_;
 };
 
+class CustomPhraseCandidateWord
+    : public CandidateWord,
+      public PinyinAbstractExtraCandidateWordInterface {
+public:
+    CustomPhraseCandidateWord(const PinyinEngine *engine, int order,
+                              std::string value)
+        : PinyinAbstractExtraCandidateWordInterface(*this, order),
+          engine_(engine) {
+        setText(Text(std::move(value)));
+    }
+
+    void select(InputContext *inputContext) const override {
+        inputContext->commitString(text().toString());
+        engine_->doReset(inputContext);
+    }
+
+private:
+    const PinyinEngine *engine_;
+};
+
+class PinyinPunctuationCandidateWord : public CandidateWord {
+public:
+    PinyinPunctuationCandidateWord(const PinyinEngine *engine, std::string word,
+                                   bool isHalf)
+        : CandidateWord(), engine_(engine), word_(std::move(word)) {
+        Text text;
+        if (isHalf) {
+            text.append(fmt::format(_("{0} (Half)"), word_));
+        } else {
+            text.append(word_);
+        }
+        setText(text);
+    }
+
+    void select(InputContext *inputContext) const override {
+        inputContext->commitString(word_);
+        engine_->doReset(inputContext);
+    }
+
+    const std::string &word() const { return word_; }
+
+private:
+    const PinyinEngine *engine_;
+    std::string word_;
+};
+
 class StrokeFilterCandidateWord : public CandidateWord {
 public:
     StrokeFilterCandidateWord(PinyinEngine *engine, Text text, int index)
@@ -215,10 +295,12 @@ private:
     std::string word_;
 };
 
-class SpellCandidateWord : public CandidateWord {
+class SpellCandidateWord : public CandidateWord,
+                           public PinyinAbstractExtraCandidateWordInterface {
 public:
-    SpellCandidateWord(PinyinEngine *engine, std::string word)
-        : engine_(engine), word_(std::move(word)) {
+    SpellCandidateWord(PinyinEngine *engine, std::string word, int order)
+        : PinyinAbstractExtraCandidateWordInterface(*this, order),
+          engine_(engine), word_(std::move(word)) {
         setText(Text(word_));
     }
 
@@ -253,20 +335,22 @@ public:
     size_t idx_;
 };
 
-class CustomCloudPinyinCandidateWord : public CloudPinyinCandidateWord {
+class CustomCloudPinyinCandidateWord
+    : public CloudPinyinCandidateWord,
+      public PinyinAbstractExtraCandidateWordInterface {
 public:
     CustomCloudPinyinCandidateWord(AddonInstance *cloudpinyin,
                                    const std::string &pinyin,
                                    const std::string &selectedSentence,
                                    InputContext *inputContext,
                                    CloudPinyinSelectedCallback callback,
-                                   bool isFirst)
+                                   int order)
         : CloudPinyinCandidateWord(cloudpinyin, pinyin, selectedSentence,
                                    inputContext, callback),
-          isFirst_(isFirst) {}
+          PinyinAbstractExtraCandidateWordInterface(*this, order) {}
 
     void select(InputContext *inputContext) const override {
-        if ((!filled() || word().empty()) && isFirst_) {
+        if ((!filled() || word().empty()) && order() == 0) {
             auto candidateList = inputContext->inputPanel().candidateList();
             for (int i = 0; i < candidateList->size(); i++) {
                 if (&candidateList->candidate(i) != this) {
@@ -276,9 +360,6 @@ public:
         }
         CloudPinyinCandidateWord::select(inputContext);
     }
-
-private:
-    bool isFirst_;
 };
 
 std::unique_ptr<CandidateList>
@@ -311,7 +392,7 @@ void PinyinEngine::initPredict(InputContext *inputContext) {
         auto &inputPanel = inputContext->inputPanel();
         inputPanel.setCandidateList(std::move(candidateList));
     } else {
-        state->predictWords_.clear();
+        state->predictWords_.reset();
     }
     inputContext->updatePreedit();
     inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
@@ -321,15 +402,16 @@ void PinyinEngine::updatePredict(InputContext *inputContext) {
     inputContext->inputPanel().reset();
 
     auto *state = inputContext->propertyFor(&factory_);
+    assert(state->predictWords_.has_value());
     auto words =
-        prediction_.predict(state->predictWords_, *config_.predictionSize);
+        prediction_.predict(*state->predictWords_, *config_.predictionSize);
     if (auto candidateList = predictCandidateList(words)) {
         auto &inputPanel = inputContext->inputPanel();
         inputPanel.setCandidateList(std::move(candidateList));
     } else {
         // Clear if we can't do predict.
         // This help other code to detect whether we are in predict.
-        state->predictWords_.clear();
+        state->predictWords_.reset();
     }
     inputContext->updatePreedit();
     inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
@@ -465,6 +547,53 @@ void PinyinEngine::updatePreedit(InputContext *inputContext) const {
     }
 }
 
+void PinyinEngine::updatePuncPreedit(InputContext *inputContext) const {
+    auto candidateList = inputContext->inputPanel().candidateList();
+
+    if (inputContext->capabilityFlags().test(CapabilityFlag::Preedit)) {
+        if (candidateList->cursorIndex() >= 0) {
+            Text preedit;
+
+            auto &candidate =
+                candidateList->candidate(candidateList->cursorIndex());
+            if (auto *puncCandidate =
+                    dynamic_cast<const PinyinPunctuationCandidateWord *>(
+                        &candidate)) {
+                preedit.append(puncCandidate->word());
+            }
+
+            preedit.setCursor(preedit.textLength());
+            inputContext->inputPanel().setClientPreedit(preedit);
+        }
+        inputContext->updatePreedit();
+    }
+}
+
+void PinyinEngine::updatePuncCandidate(
+    InputContext *inputContext, const std::string &original,
+    const std::vector<std::string> &candidates) const {
+    inputContext->inputPanel().reset();
+    auto *state = inputContext->propertyFor(&factory_);
+    auto puncCandidateList = std::make_unique<CommonCandidateList>();
+    puncCandidateList->setPageSize(*config_.pageSize);
+    puncCandidateList->setCursorPositionAfterPaging(
+        CursorPositionAfterPaging::ResetToFirst);
+    for (const auto &result : candidates) {
+        puncCandidateList->append<PinyinPunctuationCandidateWord>(
+            this, result, original == result);
+    }
+    puncCandidateList->setCursorIncludeUnselected(false);
+    puncCandidateList->setCursorKeepInSamePage(false);
+    puncCandidateList->setCursorPositionAfterPaging(
+        CursorPositionAfterPaging::SameAsLast);
+    puncCandidateList->setGlobalCursorIndex(0);
+    puncCandidateList->setSelectionKey(selectionKeys_);
+    state->mode_ = PinyinMode::Punctuation;
+    inputContext->inputPanel().setCandidateList(std::move(puncCandidateList));
+    updatePuncPreedit(inputContext);
+    inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
+}
+
 void PinyinEngine::updateUI(InputContext *inputContext) {
     inputContext->inputPanel().reset();
 
@@ -512,8 +641,38 @@ void PinyinEngine::updateUI(InputContext *inputContext) {
 
         const bool fullResult = (context.cursor() == context.size() ||
                                  context.cursor() == context.selectedLength());
+
+        std::list<std::unique_ptr<PinyinAbstractExtraCandidateWordInterface>>
+            extraCandidates;
+        std::unordered_set<std::string> customCandidateSet;
+        /// Create custom phrase candidate {{{
+        do {
+            if (selectedLength > 0 || !fullResult) {
+                break;
+            }
+            auto *results = customPhrase_.lookup(context.userInput());
+            if (!results) {
+                break;
+            }
+            for (const auto &result : *results) {
+
+                auto phrase =
+                    result.evaluate([this, inputContext](std::string_view key) {
+                        return evaluateCustomPhrase(inputContext, key);
+                    });
+                if (customCandidateSet.count(phrase)) {
+                    continue;
+                }
+                customCandidateSet.insert(phrase);
+                extraCandidates.push_back(
+                    std::make_unique<CustomPhraseCandidateWord>(
+                        this, result.order() - 1, std::move(phrase)));
+            }
+        } while (0);
+        /// }}}
+
         /// Create cloud candidate. {{{
-        std::unique_ptr<CustomCloudPinyinCandidateWord> cloud;
+        std::optional<decltype(extraCandidates)::iterator> cloud;
         if (*config_.cloudPinyinEnabled && cloudpinyin() &&
             !inputContext->capabilityFlags().testAny(
                 CapabilityFlag::PasswordOrSensitive) &&
@@ -522,10 +681,18 @@ void PinyinEngine::updateUI(InputContext *inputContext) {
             auto fullPinyin = context.useShuangpin()
                                   ? context.candidateFullPinyin(0)
                                   : context.userInput().substr(selectedLength);
-            cloud = std::make_unique<CustomCloudPinyinCandidateWord>(
+            auto cand = std::make_unique<CustomCloudPinyinCandidateWord>(
                 cloudpinyin(), fullPinyin, selectedSentence, inputContext,
                 std::bind(&PinyinEngine::cloudPinyinSelected, this, _1, _2, _3),
-                *config_.cloudPinyinIndex == 1);
+                *config_.cloudPinyinIndex - 1);
+            if (!cand->filled() ||
+                (!cand->word().empty() &&
+                 !customCandidateSet.count(cand->word()) &&
+                 !context.candidatesToCursorSet().count(cand->word()))) {
+                customCandidateSet.insert(cand->word());
+                extraCandidates.push_back(std::move(cand));
+                cloud = std::prev(extraCandidates.end());
+            }
         }
         /// }}}
 
@@ -533,33 +700,28 @@ void PinyinEngine::updateUI(InputContext *inputContext) {
         int engNess;
         auto parsedPy =
             state->context_.preedit(libime::PinyinPreeditMode::RawText);
-        std::vector<std::unique_ptr<SpellCandidateWord>> spellCands;
         if (*config_.spellEnabled && spell() && fullResult &&
             (engNess = englishNess(parsedPy, context.useShuangpin()))) {
             auto py = context.userInput().substr(selectedLength);
             auto results = spell()->call<ISpell::hintWithProvider>(
                 "en", SpellProvider::Custom, py, engNess);
-            spellCands.reserve(results.size());
             std::string bestSentence;
             if (!candidates.empty()) {
                 bestSentence = candidates[0].toString();
             }
+            int position = 1;
             for (auto &result : results) {
-                if (cloud && cloud->filled() && cloud->word() == result) {
-                    cloud.reset();
-                }
-                // Pinyin can only return eng when there is no valid option.
-                if (result == bestSentence) {
+                if (customCandidateSet.count(result) ||
+                    context.candidatesToCursorSet().count(result)) {
                     continue;
                 }
-                spellCands.push_back(
-                    std::make_unique<SpellCandidateWord>(this, result));
+                extraCandidates.push_back(std::make_unique<SpellCandidateWord>(
+                    this, result, position++));
             }
         }
         /// }}}
 
         /// Create stroke candidate {{{
-        std::vector<std::unique_ptr<StrokeCandidateWord>> strokeCands;
         if (pinyinhelper() && context.selectedLength() == 0 &&
             isStroke(context.userInput())) {
             int limit = (context.userInput().size() + 4) / 5;
@@ -568,21 +730,62 @@ void PinyinEngine::updateUI(InputContext *inputContext) {
             }
             auto results = pinyinhelper()->call<IPinyinHelper::lookupStroke>(
                 context.userInput(), limit);
+
+            int strokeCandsPos =
+                *config_.pageSize - static_cast<int>(results.size());
+            if (strokeCandsPos < 0) {
+                strokeCandsPos = *config_.pageSize - 1;
+            }
             for (auto &result : results) {
                 utf8::getChar(result.first);
                 auto py = pinyinhelper()->call<IPinyinHelper::lookup>(
                     utf8::getChar(result.first));
                 auto pystr = stringutils::join(py, " ");
-                strokeCands.push_back(std::make_unique<StrokeCandidateWord>(
-                    this, result.first, pystr));
+                extraCandidates.push_back(std::make_unique<StrokeCandidateWord>(
+                    this, result.first, pystr, strokeCandsPos++));
             }
         }
         /// }}}
 
-        size_t idx = 0;
-        for (const auto &candidate : candidates) {
+        // We expect stable sort here.
+        extraCandidates.sort([](const auto &lhs, const auto &rhs) {
+            return lhs->order() < rhs->order();
+        });
+
+        auto maybeApplyExtraCandidates = [&extraCandidates, &candidateList,
+                                          &cloud, this](bool force) {
+            if (extraCandidates.empty()) {
+                return;
+            }
+            if (candidateList->totalSize() > extraCandidates.back()->order() ||
+                candidateList->totalSize() > 2 * (*config_.pageSize) || force) {
+                // Since we will insert cloud pinyin, reset the iterator.
+                cloud.reset();
+                int lastPos = -1;
+                for (auto &extraCandidate : extraCandidates) {
+                    int actualPos = (lastPos >= extraCandidate->order())
+                                        ? lastPos
+                                        : extraCandidate->order();
+                    if (actualPos > candidateList->totalSize()) {
+                        actualPos = candidateList->totalSize();
+                    }
+                    // Rewrap it into CandidateWord.
+                    std::unique_ptr<CandidateWord> cand(
+                        &extraCandidate.release()->toCandidateWord());
+                    candidateList->insert(actualPos, std::move(cand));
+                    lastPos = actualPos;
+                }
+                extraCandidates.clear();
+            }
+        };
+
+        for (size_t idx = 0, end = candidates.size(); idx < end; idx++) {
+            const auto &candidate = candidates[idx];
             auto candidateString = candidate.toString();
-            std::vector<std::string> extraCandidates;
+            if (customCandidateSet.count(candidateString)) {
+                continue;
+            }
+            std::vector<std::string> luaExtraCandidates;
 #ifdef FCITX_HAS_LUA
             // To invoke lua trigger, we need "raw full sentence". Also, check
             // against nbest, otherwise single char may be invoked for too much
@@ -593,51 +796,20 @@ void PinyinEngine::updateUI(InputContext *inputContext) {
                 imeapi() && fullResult &&
                 candidate.sentence().back()->to()->index() ==
                     context.userInput().size()) {
-                extraCandidates =
+                luaExtraCandidates =
                     luaCandidateTrigger(inputContext, candidateString);
             }
 #endif
-            if (cloud && cloud->filled() && cloud->word() == candidateString) {
-                cloud.reset();
-            }
             candidateList->append<PinyinCandidateWord>(
                 this, Text(std::move(candidateString)), idx);
-            for (auto &extraCandidate : extraCandidates) {
+            for (auto &extraCandidate : luaExtraCandidates) {
                 candidateList->append<ExtraCandidateWord>(this, extraCandidate);
             }
-            idx++;
-            // We don't want to do too much comparison for cloud pinyin.
-            if (cloud && (!cloud->filled() || !cloud->word().empty()) &&
-                (static_cast<int>(idx) >
-                     std::max(*config_.nbest, *config_.cloudPinyinIndex - 1) ||
-                 idx == candidates.size())) {
-                auto desiredPos = *config_.cloudPinyinIndex - 1;
-                if (desiredPos > candidateList->totalSize()) {
-                    desiredPos = candidateList->totalSize();
-                }
-                candidateList->insert(desiredPos, std::move(cloud));
-            }
-            if (!spellCands.empty() && (idx == 1 || idx == candidates.size())) {
-                for (auto &spellCand : spellCands) {
-                    candidateList->append(std::move(spellCand));
-                }
-                spellCands.clear();
-            }
-            if (!strokeCands.empty() &&
-                (candidateList->totalSize() + 1 >= *config_.pageSize ||
-                 idx == candidates.size())) {
-                int desiredPos =
-                    *config_.pageSize - static_cast<int>(strokeCands.size());
-                if (desiredPos < 0 || desiredPos > candidateList->totalSize()) {
-                    desiredPos = candidateList->totalSize();
-                }
-                for (auto &strokeCand : strokeCands) {
-                    candidateList->insert(desiredPos, std::move(strokeCand));
-                    desiredPos += 1;
-                }
-                strokeCands.clear();
-            }
+
+            maybeApplyExtraCandidates(false);
         }
+
+        maybeApplyExtraCandidates(true);
         candidateList->setSelectionKey(selectionKeys_);
         if (candidateList->size()) {
             candidateList->setGlobalCursorIndex(0);
@@ -648,6 +820,23 @@ void PinyinEngine::updateUI(InputContext *inputContext) {
     inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
 }
 
+std::string PinyinEngine::evaluateCustomPhrase(InputContext *inputContext,
+                                               std::string_view key) {
+    auto result = CustomPhrase::builtinEvaluator(key);
+    if (!result.empty()) {
+        return result;
+    }
+    if (stringutils::startsWith(key, "lua:")) {
+        RawConfig config;
+        auto ret = imeapi()->call<ILuaAddon::invokeLuaFunction>(
+            inputContext, std::string(key.substr(4)), config);
+        if (!ret.value().empty()) {
+            return ret.value();
+        }
+    }
+    return "";
+}
+
 PinyinEngine::PinyinEngine(Instance *instance)
     : instance_(instance),
       factory_([this](InputContext &) { return new PinyinState(this); }) {
@@ -715,6 +904,7 @@ PinyinEngine::PinyinEngine(Instance *instance)
     loadBuiltInDict();
     reloadConfig();
     loadExtraDict();
+    loadCustomPhrase();
     instance_->inputContextManager().registerProperty("pinyinState", &factory_);
     KeySym syms[] = {
         FcitxKey_1, FcitxKey_2, FcitxKey_3, FcitxKey_4, FcitxKey_5,
@@ -862,6 +1052,27 @@ void PinyinEngine::loadExtraDict() {
     }
 }
 
+void PinyinEngine::loadCustomPhrase() {
+    const auto &standardPath = StandardPath::global();
+    auto file = standardPath.open(StandardPath::Type::PkgData,
+                                  "pinyin/customphrase", O_RDONLY);
+    if (!file.isValid()) {
+        customPhrase_.clear();
+        return;
+    }
+
+    try {
+        boost::iostreams::stream_buffer<
+            boost::iostreams::file_descriptor_source>
+            buffer(file.fd(),
+                   boost::iostreams::file_descriptor_flags::never_close_handle);
+        std::istream in(&buffer);
+        customPhrase_.load(in);
+    } catch (const std::exception &e) {
+        PINYIN_ERROR() << e.what();
+    }
+}
+
 void PinyinEngine::populateConfig() {
     if (*config_.firstRun) {
         config_.firstRun.setValue(false);
@@ -1074,6 +1285,62 @@ void PinyinEngine::deactivate(const fcitx::InputMethodEntry &entry,
     reset(entry, event);
 }
 
+bool PinyinEngine::handlePuncCandidate(KeyEvent &event) {
+    auto *inputContext = event.inputContext();
+    auto *state = inputContext->propertyFor(&factory_);
+    if (state->mode_ != PinyinMode::Punctuation) {
+        return false;
+    }
+    auto candidateList = inputContext->inputPanel().candidateList();
+    if (!candidateList) {
+        doReset(inputContext);
+        return false;
+    }
+    if (event.key().check(FcitxKey_BackSpace)) {
+        event.filterAndAccept();
+        doReset(inputContext);
+        return true;
+    }
+    int idx = event.key().keyListIndex(selectionKeys_);
+    if (idx == -1 && *config_.useKeypadAsSelectionKey) {
+        idx = event.key().keyListIndex(numpadSelectionKeys_);
+    }
+    if (idx >= 0) {
+        event.filterAndAccept();
+        if (idx < candidateList->size()) {
+            candidateList->candidate(idx).select(inputContext);
+        }
+        return true;
+    }
+
+    if (auto *movable = candidateList->toCursorMovable()) {
+        if (event.key().checkKeyList(*config_.nextCandidate)) {
+            movable->nextCandidate();
+            updatePuncPreedit(inputContext);
+            inputContext->updateUserInterface(
+                UserInterfaceComponent::InputPanel);
+            event.filterAndAccept();
+            return true;
+        }
+        if (event.key().checkKeyList(*config_.prevCandidate)) {
+            movable->prevCandidate();
+            updatePuncPreedit(inputContext);
+            inputContext->updateUserInterface(
+                UserInterfaceComponent::InputPanel);
+            event.filterAndAccept();
+            return true;
+        }
+    }
+
+    auto index = candidateList->cursorIndex();
+    if (index >= 0) {
+        candidateList->candidate(index).select(inputContext);
+    }
+
+    doReset(inputContext);
+    return false;
+}
+
 bool PinyinEngine::handleCloudpinyinTrigger(KeyEvent &event) {
     if (cloudpinyin() && event.key().checkKeyList(
                              cloudpinyin()->call<ICloudPinyin::toggleKey>())) {
@@ -1186,7 +1453,7 @@ bool PinyinEngine::handleCandidateList(KeyEvent &event) {
     auto *state = inputContext->propertyFor(&factory_);
     if ((event.key().check(FcitxKey_space) ||
          event.key().check(FcitxKey_KP_Space)) &&
-        state->predictWords_.empty()) {
+        !state->predictWords_) {
         if (candidateList->size()) {
             event.filterAndAccept();
             int idx = candidateList->cursorIndex();
@@ -1220,10 +1487,7 @@ bool PinyinEngine::handleCandidateList(KeyEvent &event) {
         }
     }
 
-    if (event.key().checkKeyList(*config_.nextPage)) {
-        event.filterAndAccept();
-        candidateList->toPageable()->next();
-        inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
+    if (handleNextPage(event)) {
         return true;
     }
 
@@ -1246,6 +1510,18 @@ bool PinyinEngine::handleCandidateList(KeyEvent &event) {
     return false;
 }
 
+bool PinyinEngine::handleNextPage(KeyEvent &event) {
+    auto *inputContext = event.inputContext();
+    auto candidateList = inputContext->inputPanel().candidateList();
+    if (event.key().checkKeyList(*config_.nextPage)) {
+        event.filterAndAccept();
+        candidateList->toPageable()->next();
+        inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
+        return true;
+    }
+    return false;
+}
+
 void PinyinEngine::updateStroke(InputContext *inputContext) {
     auto *state = inputContext->propertyFor(&factory_);
     auto &inputPanel = inputContext->inputPanel();
@@ -1340,7 +1616,7 @@ void PinyinEngine::updateForgetCandidate(InputContext *inputContext) {
     inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
 }
 
-void PinyinEngine::resetStroke(InputContext *inputContext) {
+void PinyinEngine::resetStroke(InputContext *inputContext) const {
     auto *state = inputContext->propertyFor(&factory_);
     state->strokeCandidateList_.reset();
     state->strokeBuffer_.clear();
@@ -1349,7 +1625,7 @@ void PinyinEngine::resetStroke(InputContext *inputContext) {
     }
 }
 
-void PinyinEngine::resetForgetCandidate(InputContext *inputContext) {
+void PinyinEngine::resetForgetCandidate(InputContext *inputContext) const {
     auto *state = inputContext->propertyFor(&factory_);
     state->forgetCandidateList_.reset();
     if (state->mode_ == PinyinMode::ForgetCandidate) {
@@ -1369,6 +1645,8 @@ bool PinyinEngine::handleStrokeFilter(KeyEvent &event) {
             state->strokeCandidateList_ = candidateList;
             state->mode_ = PinyinMode::StrokeFilter;
             updateStroke(inputContext);
+            handleNextPage(event);
+
             event.filterAndAccept();
             return true;
         }
@@ -1380,6 +1658,21 @@ bool PinyinEngine::handleStrokeFilter(KeyEvent &event) {
     }
 
     event.filterAndAccept();
+    // A special case that allow prev page to quit stroke filtering.
+    if ((state->strokeBuffer_.empty() &&
+         event.key().checkKeyList(*config_.prevPage))) {
+        auto candidateList = inputContext->inputPanel().candidateList();
+        if (candidateList && candidateList->toPageable() &&
+            candidateList->toPageable()->currentPage() <= 1) {
+            resetStroke(inputContext);
+            updateUI(inputContext);
+            return true;
+        }
+    }
+
+    if (handleCandidateList(event)) {
+        return true;
+    }
     // Skip all key combination.
     if (event.key().states().testAny(KeyState::SimpleMask)) {
         return true;
@@ -1432,8 +1725,8 @@ bool PinyinEngine::handleForgetCandidate(KeyEvent &event) {
     auto candidateList = inputContext->inputPanel().candidateList();
     auto *state = inputContext->propertyFor(&factory_);
     if (state->mode_ == PinyinMode::Normal) {
-        if (state->predictWords_.empty() && candidateList &&
-            candidateList->size() && candidateList->toBulk() &&
+        if (!state->predictWords_ && candidateList && candidateList->size() &&
+            candidateList->toBulk() &&
             event.key().checkKeyList(*config_.forgetWord)) {
             resetForgetCandidate(inputContext);
             state->forgetCandidateList_ = candidateList;
@@ -1489,9 +1782,18 @@ bool PinyinEngine::handlePunc(KeyEvent &event) {
     std::string punc, puncAfter;
     // skip key pad
     if (c && !event.key().isKeyPad()) {
-        std::tie(punc, puncAfter) =
-            punctuation()->call<IPunctuation::pushPunctuationV2>(
-                "zh_CN", inputContext, c);
+        auto candidates =
+            punctuation()->call<IPunctuation::getPunctuationCandidates>("zh_CN",
+                                                                        c);
+        if (candidates.size() == 1) {
+            std::tie(punc, puncAfter) =
+                punctuation()->call<IPunctuation::pushPunctuationV2>(
+                    "zh_CN", inputContext, c);
+        } else if (candidates.size() > 1) {
+            updatePuncCandidate(inputContext, utf8::UCS4ToUTF8(c), candidates);
+            event.filterAndAccept();
+            return true;
+        }
     }
     if (event.key().check(*config_.quickphraseKey) && quickphrase()) {
         auto keyString = utf8::UCS4ToUTF8(c);
@@ -1548,6 +1850,10 @@ void PinyinEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
         return;
     }
 
+    if (handlePuncCandidate(event)) {
+        return;
+    }
+
     // and by pass all modifier
     if (event.key().isModifier()) {
         return;
@@ -1561,12 +1867,12 @@ void PinyinEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
     bool lastIsPunc = state->lastIsPunc_;
     state->lastIsPunc_ = false;
 
-    // handle number key selection and prev/next page/candidate.
-    if (handleCandidateList(event)) {
+    if (handleStrokeFilter(event)) {
         return;
     }
 
-    if (handleStrokeFilter(event)) {
+    // handle number key selection and prev/next page/candidate.
+    if (handleCandidateList(event)) {
         return;
     }
 
@@ -1577,8 +1883,8 @@ void PinyinEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
     // In prediction, as long as it's not candidate selection, clear, then
     // fallback
     // to remaining operation.
-    if (!state->predictWords_.empty()) {
-        state->predictWords_.clear();
+    if (state->predictWords_) {
+        state->predictWords_.reset();
         inputContext->inputPanel().reset();
         inputContext->updatePreedit();
         inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
@@ -1791,6 +2097,8 @@ void PinyinEngine::setSubConfig(const std::string &path, const RawConfig &) {
     } else if (path == "clearalldict") {
         ime_->dict()->clear(libime::PinyinDictionary::UserDict);
         ime_->model()->history().clear();
+    } else if (path == "customphrase") {
+        loadCustomPhrase();
     }
 }
 
@@ -1799,13 +2107,13 @@ void PinyinEngine::reset(const InputMethodEntry &, InputContextEvent &event) {
     doReset(inputContext);
 }
 
-void PinyinEngine::doReset(InputContext *inputContext) {
+void PinyinEngine::doReset(InputContext *inputContext) const {
     auto *state = inputContext->propertyFor(&factory_);
     resetStroke(inputContext);
     resetForgetCandidate(inputContext);
     state->mode_ = PinyinMode::Normal;
     state->context_.clear();
-    state->predictWords_.clear();
+    state->predictWords_.reset();
     inputContext->inputPanel().reset();
     inputContext->updatePreedit();
     inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
@@ -1946,13 +2254,18 @@ void PinyinEngine::cloudPinyinSelected(InputContext *inputContext,
     auto words = state->context_.selectedWords();
     // This ensure us to convert pinyin to the right one.
     auto preedit = state->context_.preedit(libime::PinyinPreeditMode::RawText);
+    // preedit is "selected sentence" + Pinyin.
     do {
+        // Validate selected is still the same.
         if (!stringutils::startsWith(preedit, selected)) {
+            words.clear();
             break;
         }
+        // Get segmented pinyin.
         preedit = preedit.substr(selected.size());
         auto pinyins = stringutils::split(preedit, " '");
         std::string_view wordView = word;
+        // Check word length matches pinyin length.
         if (pinyins.empty() || pinyins.size() != utf8::length(word)) {
             break;
         }
@@ -2036,7 +2349,7 @@ void PinyinEngine::cloudPinyinSelected(InputContext *inputContext,
     inputContext->commitString(selected + word);
     inputContext->inputPanel().reset();
     if (*config_.predictionEnabled) {
-        state->predictWords_ = words;
+        state->predictWords_ = std::move(words);
         updatePredict(inputContext);
     }
     inputContext->updatePreedit();
