@@ -7,6 +7,7 @@
 #include "engine.h"
 #include "config.h"
 #include "context.h"
+#include "ime.h"
 #include "punctuation_public.h"
 #include "state.h"
 #include <boost/algorithm/string.hpp>
@@ -28,8 +29,13 @@
 #include <fcntl.h>
 #include <libime/core/historybigram.h>
 #include <libime/core/userlanguagemodel.h>
+#include <libime/pinyin/pinyinencoder.h>
+#include <libime/pinyin/shuangpinprofile.h>
 #include <libime/table/tablebaseddictionary.h>
+#include <map>
+#include <memory>
 #include <quickphrase_public.h>
+#include <unordered_map>
 
 namespace fcitx {
 
@@ -92,7 +98,71 @@ TableEngine::TableEngine(Instance *instance)
 
 TableEngine::~TableEngine() {}
 
-void TableEngine::reloadConfig() { readAsIni(config_, "conf/table.conf"); }
+void TableEngine::reloadConfig() {
+    readAsIni(config_, "conf/table.conf");
+    populateConfig();
+}
+
+void TableEngine::populateConfig() {
+    reverseShuangPinTable_.reset();
+
+    std::unique_ptr<libime::ShuangpinProfile> shuangpinProfile;
+
+    if (*config_.shuangpinProfile == LookupShuangpinProfileEnum::Custom) {
+        auto file = StandardPath::global().open(StandardPath::Type::PkgConfig,
+                                                "pinyin/sp.dat", O_RDONLY);
+        if (file.isValid()) {
+            try {
+                boost::iostreams::stream_buffer<
+                    boost::iostreams::file_descriptor_source>
+                    buffer(file.fd(), boost::iostreams::file_descriptor_flags::
+                                          never_close_handle);
+                std::istream in(&buffer);
+                shuangpinProfile =
+                    std::make_unique<libime::ShuangpinProfile>(in);
+            } catch (const std::exception &e) {
+                TABLE_ERROR() << e.what();
+            }
+        } else {
+            TABLE_ERROR() << "Failed to open shuangpin profile.";
+        }
+    } else {
+        libime::ShuangpinBuiltinProfile profile =
+            libime::ShuangpinBuiltinProfile::Ziranma;
+#define TRANS_SP_PROFILE(PROFILE)                                              \
+    case LookupShuangpinProfileEnum::PROFILE:                                  \
+        profile = libime::ShuangpinBuiltinProfile::PROFILE;                    \
+        break;
+        switch (*config_.shuangpinProfile) {
+            TRANS_SP_PROFILE(Ziranma)
+            TRANS_SP_PROFILE(MS)
+            TRANS_SP_PROFILE(Ziguang)
+            TRANS_SP_PROFILE(ABC)
+            TRANS_SP_PROFILE(Zhongwenzhixing)
+            TRANS_SP_PROFILE(PinyinJiajia)
+            TRANS_SP_PROFILE(Xiaohe)
+        case LookupShuangpinProfileEnum::No:
+        default:
+            break;
+        }
+        shuangpinProfile = std::make_unique<libime::ShuangpinProfile>(profile);
+    }
+
+    if (!shuangpinProfile) {
+        return;
+    }
+
+    reverseShuangPinTable_ =
+        std::make_unique<std::multimap<std::string, std::string>>();
+    for (const auto &[input, pys] : shuangpinProfile->table()) {
+        for (const auto &[syl, fuzzy] : pys) {
+            if (fuzzy != libime::PinyinFuzzyFlag::None) {
+                continue;
+            }
+            reverseShuangPinTable_->emplace(syl.toString(), input);
+        }
+    }
+}
 
 void TableEngine::activate(const fcitx::InputMethodEntry &entry,
                            fcitx::InputContextEvent &event) {
